// wersja SUKCES ma rysach

#define MAX_ASYNC_DEPTH_PREINIT 15
#define MAX_ITER_NEWTON_PREINIT 15

// when the size of the box is smaller than BIAS_PREINIT then Krawczyk method is run
#define BIAS_PREINIT 2 * 1e-2 // kiedys bylo 1e-2, po nowym cuteCalc w 2D lepiej dziala dla 0.05
//#define BIAS_PREINIT 10 // zawsze liczy Krawczyka
#define BIAS_CUTE_PREINIT 1e-3
// #define BIAS_CUTE_PREINIT  0
// biasCute = 0 --- always use cuteCalculation
// biasCute = 100 --- never

// when the size of the box is smaller than EPS_PREINIT then the box is declared undecided
// #define EPS_PREINIT 7.0 * 1e-3;
#define EPS_PREINIT 1e-5; // the proof from our paper on planar cc with equal masses go through for 1e-5

#include <iostream>
#include <cstdio>
#include <list>
#include <time.h>
#include <map>
#include <string>
#include <chrono>
#include <ctime>

#include <capd/vectalg/Vector_inline.h>
#include <vector>
#include <future>
#include <thread>
#include <math.h>

#include "defo.h"
#include "cc-nb.h"
#include "solutions-tests.h"
#include "test-and-help-functions.h"
#include "estimates-functions.h"
#include "my-interval-func.h"
#include "find-symmetry.h"
#include "search-alg.h"
#include "cute-calc.h"

#ifdef TESTY
#include "testy.h"
#endif

Phase phase; // indicates the phase of program: SEARCHING or TESTING

// File for binary output data
//ofstream outFileFailed;
// ofstream outFileTest;
// Files for Mathematica-format output data
ofstream mathFile;
ofstream mathListFile;
ofstream mathCyclesFile;

const int lenFileNames = 80;

char zerosFile[lenFileNames];           // the name for the file where all boxes coinating zeros (i.e. CC) found in the search stage are written - binary file
char undecidedFile[lenFileNames];       // the name for the file where all undecided boxes in the search stage or the test stage are written - binary file
char zerosTestedFile[lenFileNames];     // the name for the file where all boxes coinating zeros (i.e. CC) found in the testing stage are written - binary file
char undecidedTestFile[lenFileNames];   // the name for the file where all undecided boxes in the test stage are written - binary file
char foutfMath[lenFileNames];           // the name of the file for some mathematica data - text file
char foutfMathList[lenFileNames];       // the name of the file for some mathematica data - text file
char foutfCyclesMathList[lenFileNames]; // the name of the file for some mathematica data - text file

std::atomic<int> cnt(0); // cnt*2 --- threads' counter

atomic<long long> boxCount(0);       // number of checked boxes
atomic<long long> aprioriCounter(0); // tell how many times <checkAproriBounds failed
atomic<long long> clusterCounter(0);
atomic<long long> distanceCounter(0);
atomic<long long> checkZeroCounter(0); // checkZero 0 \in f_i
atomic<long long> KrawczykFailedCounter(0);
atomic<long long> KrawczykZeroCounter(0);
atomic<long long> KrawczykNoZeroCounter(0);
atomic<long long> UEqICounter(0); //             checkUEqI  (I=U)
atomic<long long> clusterImprovementCounter(0);
atomic<long long> interQuadCounter(0);
atomic<long long> methodFailedCounter(0);
//atomic<long long> methodZeroes(0);
atomic<long long> isoBoxesCount(0); // used in FINITENESS_TEST , to count how many times the isomorphism test worked

mutex updateCounter_mutex;
mutex krawczykSection_mutex;
mutex write_mutex;
mutex pathUpdate_mutex;

bool equalMasses;
bool orderIncreasing = false;
// it makes sense only for equalMasses, N > 4
// orderIncreasing = true: x2 <= x3 <= ... <= x_{n-3} <= x_{n-1} it makes sense only for equalMasses, N > 4
// orderIncreasing = false: x2 >= ... >= x_{n-3} >= x_{n-1}
bool fixedOrderOfBodies;
// in equalMasses case this parameter has no meaning;
// otherwise this parameter tells whether to permute bodies (fixedOrderOfBodies = false) or not (fixedOrderOfBodies = true)
// if (!fixedOrderOfBodies) then init parameters are set by aprioriBoundsData, this part of entry parameters is ignored

ZeroSearchType methodUsed = ZeroSearchType::krawczyk;
MyDouble eps = EPS_PREINIT;            //  when in the subdivion algorithm we reeach the box  with max diam over active coordinates < eps , then we deem it undecide
MyDouble bias = BIAS_PREINIT;          // for Krawczyk method, if the size of box (max diam() coordinates) >= bias, then we do not call Krawczyk operator
MyDouble biasCute = BIAS_CUTE_PREINIT; // for cuteCalculation: 0 -- always use cuteCalculation, ... 100 -- "never" use cuteCalculation
                                       // najlepszy eksperymentalnie dla N=4 to biasCute = 1e-4
                                       // dla N = 5 --- biasCute = 1e-3
                                       // dla N = 6 --- biasCute = 1e-4

int MAX_ITER_NEWTON = MAX_ITER_NEWTON_PREINIT;

int MAX_ASYNC_DEPTH = MAX_ASYNC_DEPTH_PREINIT;

// #include "testCuteCalc.cpp"  // some  tests of cuteCalculation function

int initParameters(const char *inputFile,
                   vector<MyDouble> &xInit,
                   vector<MyDouble> &yInit,
#ifdef SPATIAL
                   vector<MyDouble> &zInit,
#endif
                   vector<MyDouble> &massInit)
//  Returns: number of bodies
// structure of the data file:
// * 1st line
// 0 -- different masses, full configuration space;
//      positions in the data file are ignored
// 1 -- equal masses, bodies positions within prescribed bounds, equalMasses=true
// 2 -- different masses, bodies positions within prescribed bounds , fixedOrderOfBodies=true
// * 2nd line
// number of bodies
// * next lines
//  each line contains: intervals for x, y, z-coordinates (z in SPATIAL case) of bodies
//  and (in the case fo unequal masses) interval for masses
{
    int numBodies;

    char trash[240];

    ifstream plik(inputFile);
    if (!plik)
    {
        cout << "cannot open " << inputFile << endl;
        exit(1);
    }

    xInit.clear();
    yInit.clear();
#ifdef SPATIAL
    zInit.clear();
#endif
    massInit.clear();

    if (plik.good() == true)
    {
        int massParam;
        plik >> massParam;
        plik.getline(trash, 240); // to eat endl - and the eventual comment following

        if ((massParam < 0) || (massParam > 2))
        {
            cout << "Wrong input data -- mass parameter in the first line!" << endl;
            exit(3);
        }

        equalMasses = (massParam == 1);
        fixedOrderOfBodies = (massParam == 2);

        plik >> numBodies;
        plik.getline(trash, 240);

        if (equalMasses)
        {
            MyInterval m = MyInterval(1.0) / numBodies;
            for (int i = 0; i < numBodies - 1; i++)
            {
                massInit.push_back(m.leftBound());
                massInit.push_back(m.rightBound());
            }
        }

        // used only in the case fo unequal masses
        MyInterval massSum = MyInterval(0.0);
        MyInterval massOfBodies[numBodies];

        //-------------------------

        //  we read x,y,z-coordinates and mass
        for (int i = 0; i < numBodies - 1; i++)
        {
            if (plik.eof())
            {
                cout << "Wrong input data -- x!" << endl;
                exit(3);
            }
            MyDouble temp;
            plik >> temp;
            xInit.push_back(temp);
            plik >> temp;
            xInit.push_back(temp);

            plik >> temp;
            yInit.push_back(temp);
            plik >> temp;
            yInit.push_back(temp);

#ifdef SPATIAL
            plik >> temp;
            zInit.push_back(temp);
            plik >> temp;
            zInit.push_back(temp);
#endif
            if (!equalMasses)
            {
                MyDouble temp1, temp2;
                plik >> temp1;
                plik >> temp2;
                massOfBodies[i] = MyInterval(temp1, temp2);
                massSum += MyInterval(temp1, temp2);
            }
            plik.getline(trash, 240);
        }

        if (!equalMasses)
        {

            MyInterval lastMass = 1.0 - massSum;
            massOfBodies[numBodies - 1] = lastMass;

            // loading massOfBodies to massInit
            for (int j = 0; j < numBodies; j++)
            {
                MyDouble temp;
                temp = massOfBodies[j].leftBound();
                massInit.push_back(temp);
                temp = massOfBodies[j].rightBound();
                massInit.push_back(temp);
            }
        }

        plik.close();
    }
    else
    {
        cout << "No access to the input data file!" << std::endl;
        exit(2);
    }

    return numBodies;
}

void writeReportCounters(ostream &fout, bool nonDegeneracyTest)
// writes all counters to fout
{

    fout << endl;
    if (!nonDegeneracyTest)
        fout << "The number of zeros in the method: " << KrawczykZeroCounter << endl;
    else
        fout << "The upper bound of zeros in the method: " << isoBoxesCount << endl;

    if (methodFailedCounter > 0)
        fout << "The number of UNDECIDED cubes: " << methodFailedCounter << endl;

    fout << endl;

    fout << "Tests usage:" << endl;
    fout << "checkAprioriBounds -- " << aprioriCounter << endl;
    fout << "clusterTest -- " << clusterCounter << endl;
    fout << "distanceTest -- " << distanceCounter << endl;
    fout << "checkZero -- " << checkZeroCounter << endl;
    fout << "krawczyk: methodFailed  -- " << KrawczykFailedCounter << endl;
    fout << "krawczyk: zeroIside -- " << KrawczykZeroCounter << endl;
    fout << "krawczyk: no zero inside -- " << KrawczykNoZeroCounter << endl;
    fout << "U = I -- " << UEqICounter << endl;
    fout << "improvement in clusterTest -- " << clusterImprovementCounter << endl;
    fout << "improvement in U = I Test -- " << interQuadCounter << endl;
    if (nonDegeneracyTest)
        fout << "number of nondegenerate boxes -- " << isoBoxesCount << endl;

    fout << endl
         << "Number of checked boxes = " << boxCount << endl;

    fout << endl
         << flush;
}

void ZeroAllCounters()
{
    cnt = 0;
    boxCount = 0; // number of checked boxes

    aprioriCounter = 0; // tell how many times <checkAproriBounds failed
    clusterCounter = 0;
    distanceCounter = 0;
    checkZeroCounter = 0; // checkZero 0 \in f_i
    KrawczykFailedCounter = 0;
    KrawczykZeroCounter = 0; // 10.10.2020 utozsamilam z methodZeroes
    KrawczykNoZeroCounter = 0;
    UEqICounter = 0; //             checkUEqI  (I=U)
    clusterImprovementCounter = 0;
    interQuadCounter = 0;
    methodFailedCounter = 0;
    //methodZeroes = 0;
    isoBoxesCount = 0;
}

void writeGlobalSettings(ostream &fout, int NUM_BODIES)
{

#ifdef FINITENESS_TEST
    fout << "FINITENESS TEST " << endl;
#endif

#ifdef SPATIAL
    fout << "SPATIAL CASE" << endl;
#else
    fout << "PLANAR CASE" << endl
         << endl;
#endif
    fout << "===============================================" << endl;
    fout << "Number of bodies = " << NUM_BODIES << endl;
    if (equalMasses)
        fout << "EQUAL MASSES" << endl;
    else if (fixedOrderOfBodies)
        fout << "UNEQUAL MASSES and FIXED ORDER OF BODIES!" << endl;
    else
        fout << "UNEQUAL MASSES --- we look for ALL CC "
             << " (computations with permutations of bodies)!" << endl;
    fout << endl;
    fout << "Accuracy eps = " << eps;
    fout << ", bias = " << bias;
    fout << ", biasCute = " << biasCute << endl;
    fout << "MAX_ASYNCH_DEPTH = " << MAX_ASYNC_DEPTH << endl;
    fout << "----------------------------------------------" << endl;
    fout << endl
         << endl
         << flush;
}

void writeUndecidedBoxes(int NumBodies, const char *undecidedBoxesFile)
// the function gives report on undecided boxes
//  and also writes in Mathematica format to files 'mathFile','mathListFile' which are globally defined ostream's
{
    vector<Bodies> undecidedBoxes;
    int ccNo;
    if ((ccNo = readBodiesFromBinaryFile(NumBodies, undecidedBoxesFile, undecidedBoxes)) > 0)
    {
        cout << "the number of undecided boxes from the binary file " << ccNo << endl
             << flush;
        mathListFile << "------- undecided ----------" << endl;
        for (int k = 0; k < ccNo; k++)
        {
            undecidedBoxes[k].writeCCformatMath(mathFile, mathListFile);
            mathListFile << "-----------------" << endl;

            cout << "---------------------" << endl;
            undecidedBoxes[k].printData(true); // writes distances from 0
            cout << flush;
        }
    }
}

bool basicSearchAndTest(Bodies bodies, bool doSearch, bool doTest,
                        const char *zerosFile, const char *undecidedFile,
                        const char *zerosTestedFile, const char *undecidedTestFile)
//  zerosFile - the name for the file where all boxes coinating zeros (i.e. CC) found in the search stage are written - binary file
//            it is an output from the search stage and serves as the input to the testing stage
//   undecidedFile - the name for the file where all undecided boxes in the search stage are written - binary file
//            it is an output from the search stage
// zerosTestedFile - the name for the file where all boxes coinating zeros (i.e. CC) found in the testing stage are written - binary file
//                   this is rather an auxlilary file , can be used for diagnostic purposes
//   undecidedTestFile - the name for the file where all undecided boxes in the test stage are written - binary file
//            it is an output from the test stage
//
// uses global variables: all counters, biasCute
{
    int numBodies = bodies.N;

    bool nonDegeneracyTest = false;
    if (doSearch)
    {

        phase = searchmode;
        cout << "Input data:" << endl;
        bodies.printData();
        cout << endl;

        cout << "SEARCHING PHASE!" << endl;

        ofstream outFileSearch;
        ofstream outFileFailed;
        outFileSearch.open(zerosFile, ios::binary | ios::out | ios::trunc);
        outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);

        Count result;
#ifdef FINITENESS_TEST
        nonDegeneracyTest = true;
#endif

        try
        {
#ifdef PATH_SHOW
            const std::vector<bool> path; // a path from root to leaves in the computation tree
            result = searchCC(bodies, doSearch, nonDegeneracyTest, outFileSearch, outFileFailed, path);
#endif
#ifndef PATH_SHOW
            const std::vector<bool> path; // a path from root to leaves in the computation tree
            result = searchCC(bodies, doSearch, nonDegeneracyTest, outFileSearch, outFileFailed);
#endif
            //  solutions (boxes) found are written to   <outFileSearch>, the undecided boxes to <outFileFailed>
            // in binary format
        }
        catch (...)
        {
            cout << "Exception!" << endl;
            exit(1);
        }

        isoBoxesCount = result.nonDegenerate; // Number of nondegenerate boxes
        writeReportCounters(cout, nonDegeneracyTest);

        //#ifdef FINITENESS_TEST
        if (nonDegeneracyTest)
        {
            cout << "---------------------------------------------------------" << endl
                 << endl;

            outFileFailed.close();
            outFileSearch.close();
            outFileSearch.open(zerosFile, ios::binary | ios::out | ios::trunc);
            outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);
            ZeroAllCounters();

            if (result.nonDegenerate > 0) // jesli sa jakies niezdegonerowane
            {

                ///* //30.04.2020
                // USTALAM MASY

                Bodies *tempMidMassesBodies = bodies.calculateMidMasses();
                Bodies midMassesBodies(numBodies, tempMidMassesBodies->data); // chodzi o przeliczenie masy ostatniego ciala
                delete tempMidMassesBodies;

                cout << "FINITENESS_TEST: szukamy zer dla srodkowych mas:" << endl;
                midMassesBodies.printData();

                bool locNonDegeneracyTest = false;
                result = searchCC(midMassesBodies, doSearch, locNonDegeneracyTest, outFileSearch, outFileFailed);
                // 11.05.2020
                writeReportCounters(cout, locNonDegeneracyTest);
                cout << "---------------------------------------------------------" << endl
                     << endl;
                // 11.05.2020

                //#endif
            }
        }

        outFileFailed.close();
        outFileSearch.close();
    }

    if (doTest)
    {
        MAX_ASYNC_DEPTH = 0;
        cout << "TESTING PHASE!" << endl;
        phase = testing; // global variable
        nonDegeneracyTest = false;
        // we define some trash variables for compatybility with the call of searchCC
        //    ofstream outFileTest;
        //    ofstream outFileUndecided;
        //    outFileTest.open(zerosTestedFileName, ios::binary | ios::out | ios::trunc);
        //    outFileUndecided.open(undecidedFileName, ios::binary | ios::out | ios::trunc);
        // trash-end

        ofstream outFileTest;
        ofstream outFileUndecided;
        outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
        outFileUndecided.open(undecidedTestFile, ios::binary | ios::out | ios::trunc);

        MyDouble oldbiasCute = biasCute;
        biasCute = 0.0; // always do cuteCalculation - optimised computation of (x,y,z)^a/r^b

        int uniqueSol = 0; // the number of different solutions

        int ccNo;
        vector<Bodies> ccSolutions;
        vector<int> markedSolutions;

        // testowanie undecided --- POCZATEK 16.03.2020
        ofstream outFileSearch;

        vector<Bodies> ccUndecided;
        vector<int> markedUndecided;
        if ((ccNo = readBodiesFromBinaryFile(numBodies, undecidedFile, ccUndecided)) > 0)
        {
            cout << "START to process undecided..." << endl;
            ///*
            for (int i = 0; i < ccNo; i++)
            {
                Bodies testUnBodies(ccUndecided[i]);

                // we define some trash variables for compatybility with the call of searchCC
                //    ofstream outFileTest;
                //    ofstream outFileFailed;
                //    outFileTest.open(zerosTestedFileName, ios::binary | ios::out | ios::trunc);
                //    outFileFailed.open(undecidedFileName, ios::binary | ios::out | ios::trunc);
                // trash-end

                nonDegeneracyTest = false;
                // below in searchCC second parameter is TRUE, since we are in search stage in TESTING PHASE
                Count cZeros = searchCC(testUnBodies, true, nonDegeneracyTest, outFileTest, outFileUndecided);

                if (cZeros.zeros == 1)
                {
                    cout << "undecided no " << i << " -- CC found!" << endl;
                    outFileSearch.open(zerosFile, ios::binary | ios::out | ios::app);
                    testUnBodies.writeBinaryData(outFileSearch); // !
                    outFileSearch.close();
                }
                else
                {
                    int whatBlowUpDecides = blowUpSearch(testUnBodies, nonDegeneracyTest);
                    switch (whatBlowUpDecides)
                    {
                    case 0: // there is no solution in puffedBodies
                        cout << "There is no solution in puffedBodies!" << endl;
                        cout << "undecided no " << i << ": " << endl;
                        cout << endl
                             << endl;
                        break;

                    case 1: // there is unique solution in puffedBodies
                        cout << "There is unique solution in puffedBodies." << endl;
                        cout << "undecided no " << i << " -- CC found after blowUp!" << endl;
                        outFileSearch.open(zerosFile, ios::binary | ios::out | ios::app);
                        testUnBodies.writeBinaryData(outFileSearch); // !
                        outFileSearch.close();
                        break;

                    case 2:
                        cout << "puffedBodies remains undecided --- we too many times and we stop!" << endl;
                        cout << "undecided no " << i << " -- blowUp do not help!" << endl;
                        testUnBodies.printData();
                        cout << endl
                             << endl;
                        break;

                    case 3:
                        cout << "something wrong(!) result.zeros > 0 in blowUp!!!" << endl;
                        cout << "undecided no " << i << " -- blowUp do not help!" << endl;
                        testUnBodies.printData();
                        cout << endl
                             << endl;
                        break;

                    default:
                        break;
                    }
                }
            }
            cout << "All undecided processed!" << endl
                 << endl;
        }

        // testowanie undecided --- KONIEC

        if ((ccNo = readBodiesFromBinaryFile(numBodies, zerosFile, ccSolutions)) > 0)
        {
            /*
            for (int i = 0; i < ccNo; i++)
            {
                cout << "no " << i << ":" << endl;
                ccSolutions[i].printData();
                cout << "srednice" << endl;
                ccSolutions[i].printDiam();
            }
*/
            identifyDifferentSolutions(equalMasses, ccNo, ccSolutions, markedSolutions, uniqueSol, outFileTest, outFileUndecided); // different solutions are returned
                                                                                                                                   // as those indices i, such that markedSolutions[i]==0
            if (uniqueSol < 1)                                                                                                     // all solutions do not satisfy our order conditions
            {
                cout << endl;
                cout << "Ups!!! distanceTest failed!!!" << endl;
                // we report these solution for possible inspection
                for (int k = 0; k < ccNo; k++)
                {
                    ccSolutions[k].writeCCformatMath(mathFile, mathListFile);
                    cout << "---------------------" << endl;
                    ccSolutions[k].printData(true); // writes distances from 0
                    cout << flush;
                }
            }
            writeDiffCCandTestSymmetry(ccNo, ccSolutions, markedSolutions,
                                       outFileTest, outFileUndecided); // writes to output file different CC with the full information
                                                                       // about their symmetries
        }
        else
        {
            uniqueSol = 0;
            cout << "No CC found!" << endl;
        }

        /*
        if (methodFailedCounter > 0)
        {
            writeUndecidedBoxes(numBodies, undecidedTestFile);
        }
*/

        biasCute = oldbiasCute;

        outFileTest.close();
        outFileUndecided.close();
    }
    if (methodFailedCounter > 0)
    {
        cout << endl;
        writeUndecidedBoxes(numBodies, undecidedFile);
    }

    cout << "-------------------------------------------------" << endl
         << endl;
    return true;
}

void nextPermOfIndices(int perm[], int n, int point)
// on output: perm[point] = n-1; perm[n-2] = point; perm[n-1] = n-2
//    and perm[i]=i, for i\neq point,n-2,n-1
//   we realize the following cycle point ->  n-1 ->  n-2 -> point
{
    for (int i = 0; i < n - 1; i++)
        perm[i] = i;
    perm[n - 1] = n - 2;
    perm[n - 2] = point;
    perm[point] = n - 1;
}

void searchAndTestWithPermutations(Bodies baseBodies, bool searchStage, bool testStage, const char *zerosFile,
                                   const char *undecidedFile, const char *zerosTestedFile)
// it makes sense only if !equalMasses && !fixedOrderOfBodies
{
    int numBodies = baseBodies.N;

    Bodies bodies(baseBodies);

    int perm[numBodies];
    for (int i = 0; i < numBodies; i++)
        perm[i] = i;

    for (int point = 0; point < numBodies; point++) // this loop is indexed with the body which is the furtherst from the origin
                                                    // it is executed many times in the case of not equal masses and not fixed permutation
                                                    // MYSLE ZE TAK JAK MAMY w distanceTest - to to sa niewystarczajace permutacje, bo trzeba
                                                    // jeszcze wybierac cialo q0
    {

        cout << "Current permutation: ";
        mathFile << "Current permutation of masses: " << endl;
        mathListFile << "Current permutation of masses: " << endl;
        for (int i = 0; i < numBodies; i++)
        {
            cout << perm[i] << "  ";
            mathFile << perm[i] << "   ";
            mathListFile << perm[i] << "   ";
        }
        cout << endl;
        mathFile << endl;
        mathListFile << endl;

        // ZeroAllCounters();  // we reset all counters

        basicSearchAndTest(bodies, searchStage, testStage, zerosFile, undecidedFile, zerosTestedFile, undecidedTestFile);

        nextPermOfIndices(perm, numBodies, point);
        // on output: perm[point] = n-1; perm[n-2] = point; perm[n-1] = n-2
        // it is desirable that the last body ((n-1)-th) has to be the heaviest among q(0)... q(N-3), but at the moment we do not force this
        bodies = baseBodies;
        permuteMassesOfBodies(bodies, perm);
        setAprioriBoundsData(bodies);
    }
}
//#include "testCuteCalc.cpp"

int main(int argc, char *argv[])
{
    /*
    superTestCute();
    exit(1);
*/

    initIntervalGlobals(); // initiates sqrt2, sqrt3, pi - these variables have to be initiated somewhere in main,
                           // trying to initiate them as global variables may cause "segmentation fault" on some systems

#ifdef TESTY
    createTestBodies();
#endif

    if (argc < 4)
    {
        cerr << "Wrong number of parameters!" << endl;
        exit(1);
    }

    char idi = *argv[2]; // instance id
    if (idi < '0' || idi > '9')
    {
        cerr << "Wrong instance id!" << endl;
        exit(1);
    }

    // searchStage = true -- searching mode, false -- testing mode
    // in the searching mode we write found solutions into a binary file;
    // in testing mode only  to 'cout'; in order to get the result in a text file we should redirect the output
    bool searchStage;
    bool testStage;
    // set searchStage = true, if the parameter 's' (test) is given.
    // set testStage = true, if the parameter 't' (test) is given.
    char *par = argv[1];
    searchStage = par[0] == 's';
    testStage = par[0] == 't' || par[1] == 't';

    if (argc == 5)
        MAX_ASYNC_DEPTH = atoi(argv[4]);

    char inputFile[lenFileNames]; // the name for the basic input data file - text file

#ifdef PLANAR
    sprintf(inputFile, "data/init-2D-%s-%s.txt", argv[2], argv[3]);
    sprintf(zerosFile, "data/zeros-2D-%s-%s.dat", argv[2], argv[3]);
    sprintf(undecidedFile, "data/undecided-2D-%s-%s.dat", argv[2], argv[3]);
    sprintf(foutfMath, "data/math-2D-%s-%s.txt", argv[2], argv[3]);
    sprintf(foutfMathList, "data/math-2D-%s-%s-list.txt", argv[2], argv[3]);
    sprintf(foutfCyclesMathList, "data/math-cycles-2D-%s-%s-list.txt", argv[2], argv[3]);
#endif
#ifdef SPATIAL
    sprintf(inputFile, "data/init-3D-%s-%s.txt", argv[2], argv[3]);
    sprintf(zerosFile, "data/zeros-3D-%s-%s.dat", argv[2], argv[3]);
    sprintf(undecidedFile, "data/undecided-3D-%s-%s.dat", argv[2], argv[3]);
    sprintf(foutfMath, "data/math-3D-%s-%s.txt", argv[2], argv[3]);
    sprintf(foutfMathList, "data/math-3D-%s-%s-list.txt", argv[2], argv[3]);
    sprintf(foutfCyclesMathList, "data/math-cycles-3D-%s-%s-list.txt", argv[2], argv[3]);
#endif
    sprintf(zerosTestedFile, "data/tested-zeros-%s-%s.dat", argv[2], argv[3]);
    sprintf(undecidedTestFile, "data/tested-undecided-%s-%s.dat", argv[2], argv[3]);

    mathFile.open(foutfMath, ios::out | ios::trunc);
    mathFile.setf(ios::fixed, ios::floatfield);
    mathFile.precision(10);

    mathListFile.open(foutfMathList, ios::out | ios::trunc);
    mathListFile.setf(ios::fixed, ios::floatfield);
    mathListFile.precision(10);

    mathCyclesFile.open(foutfCyclesMathList, ios::out | ios::trunc);
    mathCyclesFile.setf(ios::fixed, ios::floatfield);
    mathCyclesFile.precision(10);

    vector<MyDouble> xInit, yInit, zInit, massInit;

#ifdef SPATIAL
    int NUM_BODIES = initParameters(inputFile, xInit, yInit, zInit, massInit);
#else // PLANAR
    int NUM_BODIES = initParameters(inputFile, xInit, yInit, massInit);
#endif

    if (atoi(argv[2]) != NUM_BODIES)
    {
        cout << "Wrong input data! the numbers of bodies are incorrect  " << atoi(argv[2]) << "not equal to " << NUM_BODIES << endl;
        exit(2);
    }

    Bodies bodies(NUM_BODIES, true,
                  xInit,
                  yInit,
#ifdef SPATIAL
                  zInit,
#endif
                  massInit);

    if (!equalMasses)
        if (!fixedOrderOfBodies)
            setAprioriBoundsData(bodies);

    if (bias < eps)
        bias = eps;

#ifdef FINITENESS_TEST
    testStage = false; // we just compute the boxes wherewhere we have at most one solution, hence  the testing does not make sense
    if (testStage)
    {
        cout << "in the FINITENESS_TEST mode the testing does not make sense" << endl
             << flush;
        exit(1);
    }
#endif

    std::cout.precision(15);
    writeGlobalSettings(cout, NUM_BODIES);
    std::time_t t = std::time(nullptr);
    std::cout << std::ctime(&t);

    if (equalMasses || fixedOrderOfBodies)
    {
        basicSearchAndTest(bodies, searchStage, testStage, zerosFile, undecidedFile, zerosTestedFile, undecidedTestFile);
    }
    else
    {
        searchAndTestWithPermutations(bodies, searchStage, testStage, zerosFile, undecidedFile, zerosTestedFile);
    }
    mathFile.close();
    mathListFile.close();
    mathCyclesFile.close();
    return 0;
}
