
MyInterval radiusXYn(const MyInterval &x, const MyInterval &y, int exp)
{
  return power(sqrt(power(x, 2) + power(y, 2)), exp);
}

void setMinMax2D(const MyInterval &x, const MyInterval &y,
                 MyInterval &qx, MyInterval &qy)
{
  qx = intervalHull(x, qx);
  qy = intervalHull(y, qy);
}
void xydivr52D(const MyInterval &x, const MyInterval &y,
               MyInterval &tempxy)
// tempxy = xy/r^{5}
{
  MyInterval rMinus5 = radiusXYn(x, y, -5);
  tempxy = rMinus5 * x * y;
}

void compXYtoAoverRtoB(const MyInterval &x, const MyInterval &y,
                       MyInterval &tempx, MyInterval &tempy,
                       int rExp, int bExp)
// On output:
// tempx = x^bExp/r^rExp
// tempy = y^bExp/r^rExp
{
  MyInterval r = radiusXYn(x, y, -rExp);
  tempx = r * power(x, bExp);
  tempy = r * power(y, bExp);
}

bool intersectUppLowEdgeWithLine(const MyInterval &coef, const MyInterval &x, const MyInterval &y0,
                                 MyInterval &pointx, MyInterval &pointy)
//  line: x = y*coef
//  returns: true if the line has an intersection line y=y0 on the egde x \times {y=y_0} and the intersection is returned in (pointx,pointy)
//    otherwise false is returned
{
  MyInterval temp;
  bool intersect;

  pointy = y0;

  temp = coef * pointy;
  intersect = myIntersection(temp, x, pointx);

  return intersect;
}

bool checkValCritPoints(const MyInterval &x, const MyInterval &y, int b, int a, MyInterval &qx)
// for critical points on the edges we compute x^a/r^b
// returns true, if there is at least one such critical point

////  we assume that (0,0) does not belong to x \times y
// the critical points are given by:
//  on x_0=x.left() or x_0=x.right()  y=0
//  on y_0=y.left() or y_0=y.right()   x=\pm \sqrt{a/b-a} y_0  or (x=0 if a>1)
{
  bool retval = false;

  MyInterval r, temp;

  //  on left and right edges   x_0=x.left() or x_0=x.right()
  if (isSingular(y))
  {
    retval = true;
    //  x is either positive or negative
    r = x;
    if (x < 0)
      r = -x;
    qx = 1.0 / power(r, b - a); // because x^a/r^b= x^a/|x|^b=1/|x|^{b-a}  \times sign (x^a)
    if (power(x, a) < 0)
      qx = -qx;
  }

  // on lower and upper edges: y_0=y.left() or y_0=y.right()
  if (isSingular(x))
  { // point x=0
    if (!retval)
    {
      retval = true;
      qx = 0;
    }
    else
    {
      qx = intervalHull(qx, 0);
    }
  }

  MyInterval coef = sqrt(MyInterval(a) / (b - a));

  MyInterval pointx, pointy;
  if (intersectUppLowEdgeWithLine(coef, x, y.left(), pointx, pointy))
  {
    r = radiusXYn(pointx, pointy, b);
    temp = power(pointx, a) / r;

    if (!retval)
    {
      retval = true;
      qx = temp;
    }
    else
    {
      qx = intervalHull(qx, temp);
    }
  }

  if (intersectUppLowEdgeWithLine(-coef, x, y.left(), pointx, pointy))
  {
    r = radiusXYn(pointx, pointy, b);
    temp = power(pointx, a) / r;

    if (!retval)
    {
      retval = true;
      qx = temp;
    }
    else
    {
      qx = intervalHull(qx, temp);
    }
  }

  if (intersectUppLowEdgeWithLine(coef, x, y.right(), pointx, pointy))
  {
    r = radiusXYn(pointx, pointy, b);
    temp = power(pointx, a) / r;

    if (!retval)
    {
      retval = true;
      qx = temp;
    }
    else
    {
      qx = intervalHull(qx, temp);
    }
  }

  if (intersectUppLowEdgeWithLine(-coef, x, y.right(), pointx, pointy))
  {
    r = radiusXYn(pointx, pointy, b);
    temp = power(pointx, a) / r;

    if (!retval)
    {
      retval = true;
      qx = temp;
    }
    else
    {
      qx = intervalHull(qx, temp);
    }
  }

  return retval;
}

// 27.11 -- poczatek
/*-------------------------------------------------------------------------------------------*/
inline bool intersectLineWithEdge2D(MyInterval coef, MyInterval qx, MyInterval qy,
                                    MyInterval &point)
// check weather qx has non-empty intersection with  coef* qy

// qx is an interval (an egde), qy -- a point (yL or yR)
// coef = +/- 1/2
{
  return myIntersection(qx, coef * qy, point);
}

void computeFourPoints2D(const MyInterval &x, const MyInterval &y,
                         MyInterval &qxy)
// on output: qxy = intervalHull(values_selected_4points, qxy)
// we compute xy/r^{5}
//
// 4 points:
// intersection of the line `x = +/-1/2 * y` and the edges y =yL or y = yR of the qube `x\times y`:
{
  MyInterval tempxy;
  MyInterval xI;
  MyInterval r, rMinus5;
  MyInterval coef = MyInterval(1.0) / 2;

  MyInterval yAcc = y.right(); // upper edge
  if (intersectLineWithEdge2D(coef, x, yAcc, xI))
  {
    xydivr52D(xI, yAcc, tempxy);
    qxy = intervalHull(tempxy, qxy); // set MinMax
  }
  if (intersectLineWithEdge2D(-coef, x, yAcc, xI))
  {
    xydivr52D(xI, yAcc, tempxy);
    qxy = intervalHull(tempxy, qxy); // set MinMax
  }

  yAcc = y.left(); // lower edge
  if (intersectLineWithEdge2D(coef, x, yAcc, xI))
  {
    xydivr52D(xI, yAcc, tempxy);
    qxy = intervalHull(tempxy, qxy); // set MinMax
  }
  if (intersectLineWithEdge2D(-coef, x, yAcc, xI))
  {
    xydivr52D(xI, yAcc, tempxy);
    qxy = intervalHull(tempxy, qxy); // set MinMax
  }
}

// 29.11 -- poczatek
void sqrXDivr52D(const MyInterval &x, const MyInterval &y,
                 MyInterval &tempx, MyInterval &tempy)
// tempx = 3x^2/r^{5} - 1/r^{3}
// tempy = 3y^2/r^{5} - 1/r^{3}
{
  MyInterval rMinus5 = radiusXYn(x, y, -5);
  tempx = (2 * power(x, 2) - power(y, 2)) * rMinus5;
  tempy = (2 * power(y, 2) - power(x, 2)) * rMinus5;
}

void computeFourPointsXAndY2D(MyInterval coef, const MyInterval &x, const MyInterval &y,
                              MyInterval &qx)
// on output:
// qx = intervalHull(values_selected_4points, qx)
// we compute 3x^2/r^{5} - 1/r^{3}
//
// 4 points:
// intersection of the line `x = +/-coef * y` and the edges y =yL or y = yR of the qube `x\times y`:
{
  MyInterval tempx, tempy;
  MyInterval xI;
  MyInterval r, rMinus5;

  MyInterval yAcc = y.right(); // upper edge
  if (intersectLineWithEdge2D(coef, x, yAcc, xI))
  {
    sqrXDivr52D(xI, yAcc, tempx, tempy);
    qx = intervalHull(tempx, qx); // set MinMax
  }
  if (intersectLineWithEdge2D(-coef, x, yAcc, xI))
  {
    sqrXDivr52D(xI, yAcc, tempx, tempy);
    qx = intervalHull(tempx, qx); // set MinMax
  }

  yAcc = y.left(); // lower edge
  if (intersectLineWithEdge2D(coef, x, yAcc, xI))
  {
    sqrXDivr52D(xI, yAcc, tempx, tempy);
    qx = intervalHull(tempx, qx); // set MinMax
  }
  if (intersectLineWithEdge2D(-coef, x, yAcc, xI))
  {
    sqrXDivr52D(xI, yAcc, tempx, tempy);
    qx = intervalHull(tempx, qx); // set MinMax
  }
}

void computeXAndYInCorners2D(const MyInterval &x, const MyInterval &y,
                             MyInterval &qx, MyInterval &qy)
// On output:
// {qx,qy}=IntervalHull((3x^2/r^{5} - 1/r^{3}, (3y^2/r^{5} - 1/r^{3})
// on corners of box x \times y
//
//   4-------------1
//   |             |
//   |             |
//   3-------------2
//
{
  MyInterval tempx, tempy;
  // bottom-left corner: (1) x -- right, y -- right
  MyInterval xAcc = x.right();
  MyInterval yAcc = y.right();
  sqrXDivr52D(xAcc, yAcc, qx, qy);

  // top-left corner: (2)  x -- right, y -- left
  yAcc = y.left();
  sqrXDivr52D(xAcc, yAcc, tempx, tempy);
  qx = intervalHull(tempx, qx); // set MinMax
  qy = intervalHull(tempy, qy); // set MinMax

  // top-right corner: (3)  x -- left, y -- left
  xAcc = x.left();
  sqrXDivr52D(xAcc, yAcc, tempx, tempy);
  qx = intervalHull(tempx, qx); // set MinMax
  qy = intervalHull(tempy, qy); // set MinMax

  // bottom-right corner: (4)    x -- left, y -- right
  yAcc = y.right();
  sqrXDivr52D(xAcc, yAcc, tempx, tempy);
  qx = intervalHull(tempx, qx); // set MinMax
  qy = intervalHull(tempy, qy); // set MinMax
}

// 29.11 -- koniec

void computeXYInCorners2D(const MyInterval &x, const MyInterval &y,
                          MyInterval &qxy)
// On output: {qxy}=IntervalHull(xy/r^{5})
// on corners of box x \times y
//
//   4-------------1
//   |             |
//   |             |
//   3-------------2
//
{
  MyInterval tempxy;
  // bottom-left corner: (1) x -- right, y -- right
  MyInterval xAcc = x.right();
  MyInterval yAcc = y.right();
  xydivr52D(xAcc, yAcc, qxy);

  // top-left corner: (2)  x -- right, y -- left
  yAcc = y.left();
  xydivr52D(xAcc, yAcc, tempxy);
  qxy = intervalHull(tempxy, qxy); // set MinMax

  // top-right corner: (3)  x -- left, y -- left
  xAcc = x.left();
  xydivr52D(xAcc, yAcc, tempxy);
  qxy = intervalHull(tempxy, qxy); // set MinMax

  // bottom-right corner: (4)    x -- left, y -- right
  yAcc = y.right();
  xydivr52D(xAcc, yAcc, tempxy);
  qxy = intervalHull(tempxy, qxy); // set MinMax
}

void cuteCalcMixed(const MyInterval &x, const MyInterval &y,
                   MyInterval &qxy)
/* we are calculating:
 * qxy = xy/r^{5}
 */

/*
 * we consider lines:
 * * y = +/- 2x
 * * x = 0, y - any
 * * y = 0, x - any
 */
{
  MyInterval tempxy;

  MyInterval r = sqrt(power(x, 2) + power(y, 2));

  if (!testIfDoCute(r)) // we do a naive evaluation
  {
    MyInterval r5 = power(r, -5);
    qxy = x * y * r5;
    return;
  }

  // first check the corners of the `qi` qube:
  computeXYInCorners2D(x, y, qxy);
  // now we have: qxy -- min and max from the corners

  // 4 points:
  // intersection of the line `x = +/-1/2 * y` and the edges y =yL or y = yR of the qube `x\times y`:
  computeFourPoints2D(x, y, qxy);
  // 4 points:
  // intersection of the line `y = +/-1/2 * x` and the edges x =xL or x = xR of the qube `x\times y`:
  computeFourPoints2D(y, x, qxy);
}

// 27.11 --- koniec

// 29.11 -- poczatek

void cuteCalculationFast(const MyInterval &x, const MyInterval &y,
                         MyInterval &qx, MyInterval &qy)
/* we are calculating:
 * qx = 3x^/r^{5} - 1/r^3
 * qy = 3y^/r^{5} - 1/r^3
 */

/*
 * we consider lines:
 * * y = +/- coef*x, gdzie coef = \sqrt{3/2}
 * * y = +/- coef*x, gdzie coef = 2
 */
{
  MyInterval tempx, tempy;

  MyInterval r = sqrt(power(x, 2) + power(y, 2));

  if (!testIfDoCute(r)) // we do a naive evaluation
  {
    sqrXDivr52D(x, y, qx, qy);
    return;
  }

  // first check the corners of the `qi` qube:
  computeXAndYInCorners2D(x, y, qx, qy);
  // now we have: qxy -- min and max from the corners

  MyInterval coef = sqrt(MyInterval(3.0) / 2);
  // 4 points:
  // intersection of the line `x = +/-coef * y` and the edges y =yL or y = yR of the qube `x\times y`:
  computeFourPointsXAndY2D(coef, x, y, qx);
  // intersection of the line `y = +/-coef * x` and the edges x =xL or x = xR of the qube `x\times y`:
  computeFourPointsXAndY2D(coef, y, x, qy);

  coef = MyInterval(2.0);
  // 4 points:
  // intersection of the line `y = +/-2 * x` and the edges x =xL or x = xR of the qube `x\times y`:
  computeFourPointsXAndY2D(coef, y, x, qy);
  // intersection of the line `x = +/-2 * y` and the edges y =yL or y = yR of the qube `x\times y`:
  computeFourPointsXAndY2D(coef, y, x, qx);
}

// 29.11 -- koniec

void cuteCalculation(const MyInterval &x, const MyInterval &y, MyInterval &qx, MyInterval &qy,
                     int rExp, int bExp)
/*
 * we are calculating: q^bExp/|q|^{rExp}
 On output:
 * qx = x^bExp/r^{rExp}, qy = y^bExp/r^{rExp}
   where (x,y) \in x \times y
 */
/*
  in the paper: bExp = a, rExp = b
 * we have to check corners and intersection of lines
 * * x = +/- sqrt(b/(r-b)) * y
 * * y  = +/- sqrt(b/(r-b)) * x
 * * x = 0
 * * y = 0
 with edges

 * to find the extremal values//
 */

{
  if (rExp <= bExp)
  {
    cout << "Wrong exponents!";
    exit(1);
  }

  MyInterval r = sqrt(power(x, 2) + power(y, 2));
  MyInterval r3 = power(r, -rExp);
  if (!testIfDoCute(r)) // we do a naive evaluation
  {
    qx = power(x, bExp) * r3;
    qy = power(y, bExp) * r3;
    return;
  }

  // first we evaluate on the corners of the `x\times y` qube:
  // left-bottom corner
  MyInterval xAcc = x.left();
  MyInterval yAcc = y.left();
  MyInterval tempx;
  MyInterval tempy;
  compXYtoAoverRtoB(xAcc, yAcc, tempx, tempy, rExp, bExp);
  qx = tempx;
  qy = tempy;

  // top-left corner:
  yAcc = y.right();
  compXYtoAoverRtoB(xAcc, yAcc, tempx, tempy, rExp, bExp);
  setMinMax2D(tempx, tempy, qx, qy);

  // top-right corner:
  xAcc = x.right();
  compXYtoAoverRtoB(xAcc, yAcc, tempx, tempy, rExp, bExp);
  setMinMax2D(tempx, tempy, qx, qy);

  // bottom-right corner:
  yAcc = y.left();
  compXYtoAoverRtoB(xAcc, yAcc, tempx, tempy, rExp, bExp);
  setMinMax2D(tempx, tempy, qx, qy);

  // now we have:  min and max from the corners  in qx,qy

  if (!checkValCritPoints(x, y, rExp, bExp, tempx)) // for critical points on the edges we compute x^bexp/r^{rExp}
    tempx = qx;
  if (!checkValCritPoints(y, x, rExp, bExp, tempy)) //  the same as above but with x and y roles exchaged
    tempy = qy;

  setMinMax2D(tempx, tempy, qx, qy);
}
