#ifndef krawnet_h
#define krawnet_h

#include "defo.h"
#include "bodies.h"

enum class ResultType
{
  uniqueZero = 1,     // there is exactly one zero
  methodFailed = 2,   // the matrix is singular or Krawczyk/Newton operator result contains the initial condition
  noZeroInSet = 3,    // no zero found
  collisionFound = 4, // collision found
  twoZeros = 5,       // two zeros found
  zeroFound = 6       // zero found, but already processed
};

enum class ZeroSearchType
{
  krawczyk,
  newton
};

ResultType newtonMethod(Bodies copyBodies, Bodies &returnValues, bool &intersec, MyMatrix &DFF);



ResultType krawczykMethod(Bodies boxBodies, Bodies &returnValues, bool &intersec, MyMatrix &DFF, bool recomputeDFF , bool thisIsUniqueZero );
  // the Krawczyk method is applied to <boxBodies>
  // On input: DFF = DfRed(boxbodies) - it is supplied because this is a costly function and is used elsewhere
  //     recomputeDFF - tells whether in each iteration we compute DFF, otherwise we just use DFF supplied on input
  //     if thisIsUniqueZero == false, then it is a normal run of the Krawczyk operator , we iterate ... and check the inclusion
  //        in the interior to have the proof of the existence of the solution
  //     if thisIsUniqueZero == true, then we KNOW from somewhere that <boxBodies> contains a unique zero
  //        and we just use the Krawczyk operator in order to refine the bounds
  ////////////
  //   one reason for the parameter "thisIsUniqueZero" is the following observation ; when iterating a box with a unique solution 
  //   for very many iterates it may happen that the box stabilizes and no longer we have that Krawczyk(box) \subset int box
  //     still in such case we do not want return the message that the method failed  
  // On output:
  //    intersec = true  when the method fails but still KrawczykMethod succeded in reducing the size of the search box
  //               false otherwise
  //   returnValues - is the result of Newton/Krawczyk operator interations (Nx \cap x if the intersection happens to be non-empty)
  //    DFF contains DfRed(returnValues) if recomputeDFF==true


ResultType refineCC(Bodies &bodies);
// the Krawczyk method is applied to a box containing CC in order to refine bounds

#endif
